@@ -345,22 +345,13 @@ class ObserverHiveEmitter(Emitter):
             'countOfRunsToEmit': len(unemitted_runs)
         })
 
-
-        # result = self.repo.try_upsert(df=self.create_df(table, unemitted_runs),
-        #                               partition_puning_col=self.partition_col(),
-        #                               partition_cols=(self.partition_col(),))
-
-        result = self.repo.try_append(df=self.create_df(table, unemitted_runs),
-                                      partition_cols=(self.partition_col(),))
+        result = self.repo.try_write_append(df=self.create_df(table, unemitted_runs))
 
         if result.is_left():
             logger.info('jobsworth:observer:emit UPSERT ERROR', ctx=result.error().error())
         self.emitted_map.update(unemitted_runs)
         return result
 
-    def partition_col(self):
-        return 'hasRunDateUTC'
-
     def read(self):
         return self.repo.read()
 
