@@ -8,6 +8,10 @@ from jobsworthy import repo
 class Observer(repo.HiveRepo):
     table_name = "observer"
 
+    partition_columns = ("hasRunDateUTC",)
+
+    pruning_column = 'hasRunDateUTC'
+
     def identity_merge_condition(self, name_of_baseline, update_name):
         return f"{name_of_baseline}.run.id = {update_name}.run.id"
 
