@@ -1,6 +1,7 @@
 from .spark_db import (
     Db,
-    DbNamingConventionCallerDefined
+    DbNamingConventionCallerDefined,
+    DbNamingConventionDomainBased
 )
 
 
