@@ -5,7 +5,7 @@ import pyspark.sql.streaming
 from pyspark.sql import dataframe
 from pyspark.sql import functions as F
 from delta.tables import *
-from functools import reduce
+from pymonad.tools import curry
 
 from . import spark_db, repo_messages
 from jobsworthy.util import error, monad
@@ -117,9 +117,33 @@ class HiveRepo:
         self.stream_writer = stream_writer
         self.reader = reader
         self.stream_query = None
-        if not hasattr(self, "table_name") or not self.__class__.table_name:
+        if not self.name_of_table():
             raise repo_messages.table_name_not_configured()
 
+    #
+    # Table Properties
+    #
+    def partition_on(self):
+        return self.__class__.partition_columns if hasattr(self, 'partition_columns') else tuple()
+
+    def name_of_table(self):
+        return self.__class__.table_name if hasattr(self, "table_name") else None
+
+    def table_schema(self):
+        return self.__class__.schema if hasattr(self, "schema") else None
+
+    def prune_on(self):
+        return self.__class__.pruning_column if hasattr(self, 'pruning_column') else None
+
+    def associated_temporary_table(self):
+        return self.__class__.temp_table_name if hasattr(self, "temp_table_name") else None
+
+    def merge_condition(self):
+        return self.identity_merge_condition if hasattr(self, 'identity_merge_condition') else None
+
+    def asserted_table_properties(self):
+        return self.__class__.table_properties if hasattr(self, 'table_properties') else None
+
     #
     # Table Lifecycle Events
     #
@@ -127,6 +151,23 @@ class HiveRepo:
         table_name = table_name if table_name else self.table_name
         return self.db.table_exists(table_name)
 
+    def is_delta_table_from_path(self) -> monad.EitherMonad:
+        """
+        Useful when the table is not registered with HIVE.  Attempts to create a DF from a delta path.
+        Right means the delta table exists, Left means it doesn't.
+        :param repo:
+        :param table_name:
+        :return:
+        """
+        return monad.monadic_try()(self.delta_table)()
+
+    def create_as_unmanaged_delta_table(self):
+        return self.db.session.sql(f"""
+        CREATE TABLE {self.table_name}
+        USING DELTA 
+        LOCATION '{self.db.naming().db_table_path(self.table_name)}'
+        """)
+
     def drop_table(self):
         """
         Drops the table if it exists.  Use carefully!
@@ -146,10 +187,10 @@ class HiveRepo:
         return self
 
     def drop_temp_table(self):
-        if not hasattr(self, "temp_table_name") or not self.__class__.temp_table_name:
+        if not self.associated_temporary_table():
             raise repo_messages.temp_table_not_configured()
 
-        return self.drop_table_by_name(table_to_drop=self.temp_table_name)
+        return self.drop_table_by_name(table_to_drop=self.associated_temporary_table())
 
     #
     # Table Read Functions
@@ -183,15 +224,67 @@ class HiveRepo:
     def determine_schema_to_use(self, schema_from_argument=None):
         if schema_from_argument:
             return schema_from_argument
-        if hasattr(self, "schema") and self.schema():
-            return self.schema()
+        if self.table_schema():
+            return self.table_schema()
         raise repo_messages.no_schema_provided_on_create_df()
 
     @monad.monadic_try(error_cls=error.RepoWriteError)
-    def try_write_stream(self, stream, partition_cols: tuple = tuple(), trigger: dict = None):
-        return self.write_stream(stream, partition_cols, trigger)
+    def try_write_append(self, df):
+        return self.write_append(df)
 
-    def write_stream(self, stream, partition_cols: tuple = tuple(), trigger: dict = None):
+    def write_append(self, df):
+        """
+        Executes a simple append operation on a table using the provided dataframe.
+        + Optionally provide a tuple of columns for partitioning the table.
+        """
+        result = (df.write
+                  .format(self.db.table_format())
+                  .partitionBy(self.partition_on())
+                  .mode("append")
+                  .saveAsTable(self.db_table_name()))
+        self.merge_table_properties()
+        return result
+
+    @monad.monadic_try(error_cls=error.RepoWriteError)
+    def try_upsert(self, df):
+        """
+        The try_upsert wraps the upsert function with a Try monad.  The result will be an Either.  A successful result
+        usually returns Right(None).
+        """
+        return self.upsert(df)
+
+    def upsert(self, df):
+        """
+        Upsert performs either a create or a delta merge function.  The create is called when the table doesnt exist.
+        Otherwise a delta merge is performed.
+        + partition_pruning_col.  When partitioning, the merge will use one of the partition columns to execute a merge using
+          partition pruning.
+        + Optionally provide partition columns as a tuple of column names.
+
+        Note that the merge requires that the repository implement a identity_merge_condition function when using merge
+        operations.  This is a merge condition that identifies the upsert identity.
+        """
+        if not self.table_exists():
+            return self.write_append(df)
+
+        return self._perform_upsert(df, None)
+
+    def _perform_upsert(self, df, _batch_id=None):
+        return (self.delta_table().alias(self.table_name)
+                .merge(df.alias('updates'),
+                       self.build_merge_condition(self.table_name, 'updates', self.prune_on()))
+                .whenNotMatchedInsertAll()
+                .whenMatchedUpdateAll()
+                .execute())
+
+    #
+    # Streaming Functions
+    #
+    @monad.monadic_try(error_cls=error.RepoWriteError)
+    def try_write_stream(self, stream, trigger: dict = None):
+        return self.write_stream_append(stream, trigger)
+
+    def write_stream_append(self, stream, trigger: dict = None):
         """
         Write a stream.  Provide a stream.
         + To create partition columns in the table, provide a tuple of column names, the default is no partitioning
@@ -201,11 +294,11 @@ class HiveRepo:
         """
         trigger_condition = trigger if trigger else self.__class__.default_stream_trigger_condition
 
-        self.stream_query = self.write_stream_to_table(stream, self.table_name, trigger_condition, partition_cols)
+        self.stream_query = self._write_stream_append_only(stream, self.table_name, trigger_condition)
         return self
 
-    def write_stream_temporary(self, stream, partition_cols: tuple = tuple()) -> DataFrame:
-        if not hasattr(self, "temp_table_name") or not self.__class__.temp_table_name:
+    def write_stream_temporary(self, stream) -> DataFrame:
+        if not self.associated_temporary_table():
             raise repo_messages.temp_table_not_configured()
 
         """
@@ -218,94 +311,69 @@ class HiveRepo:
 
         self.drop_temp_table()
 
-        self.stream_query = self.write_stream_to_table(stream, self.temp_table_name, trigger_condition, partition_cols)
+        self.stream_query = self._write_stream_append_only(stream,
+                                                           self.associated_temporary_table(),
+                                                           trigger_condition)
         self.await_termination()
         df = self.read(self.temp_table_name)
         return df
 
-    def delta_stream_upserter(self,
-                              stream: dataframe.DataFrame,
-                              partition_pruning_col: str = None,
-                              partition_cols: tuple = tuple()):
-        return DeltaStreamUpserter(self, partition_pruning_col, partition_cols).execute(stream)
+    # def delta_stream_upserter(self,
+    #                           stream: dataframe.DataFrame,
+    #                           partition_pruning_col: str = None,
+    #                           partition_cols: tuple = tuple()):
+    #     return DeltaStreamUpserter(self, partition_pruning_col, partition_cols).execute(stream)
 
-    def write_stream_to_table(self,
-                              stream,
-                              table_name: str,
-                              trigger_condition,
-                              partition_cols: tuple = tuple()):
-        if not stream.isStreaming:
+    def stream_write_via_delta_upsert(self,
+                                      stream,
+                                      trigger: Dict = None):
+
+        trigger_condition = trigger if trigger else self.__class__.default_stream_trigger_condition
+
+        self._raise_when_not_in_stream(stream)
+
+        if self.is_delta_table_from_path().is_left():
+            self.stream_query = self._write_stream_append_only(stream, self.table_name, trigger_condition)
+        else:
+            self.stream_query = self._stream_write_upsert(stream, self.table_name, trigger_condition)
+        return self
+
+    def _raise_when_not_in_stream(self, df):
+        if not df.isStreaming:
             raise repo_messages.df_not_in_stream()
         if not self.stream_writer:
             raise repo_messages.writing_stream_without_setting_writer()
 
-        return self.stream_writer().write(self, stream
-                                          .writeStream
+    def _write_stream_append_only(self,
+                                  stream,
+                                  table_name: str,
+                                  trigger_condition):
+
+        self._raise_when_not_in_stream(stream)
+
+        return self.stream_writer().write(self,
+                                          stream.writeStream
                                           .format('delta')
-                                          .partitionBy(partition_cols)
+                                          .partitionBy(self.partition_on())
                                           .option('checkpointLocation',
                                                   self.db.naming().checkpoint_location(table_name))
-                                          .trigger(**trigger_condition), table_name)
+                                          .trigger(**trigger_condition),
+                                          table_name)
 
-    @monad.monadic_try(error_cls=error.RepoWriteError)
-    def try_append(self, df, partition_cols: tuple = tuple()):
-        return self.append(df, partition_cols)
+    def _stream_write_upsert(self,
+                             stream,
+                             table_name: str,
+                             trigger_condition):
+        return self.stream_writer().write(self,
+                                          stream.writeStream
+                                          .format('delta')
+                                          .option('checkpointLocation',
+                                                  self.db.naming().checkpoint_location(table_name))
+                                          .trigger(**trigger_condition)
+                                          .foreachBatch(self._perform_upsert)
+                                          .outputMode('update'),
+                                          table_name)
 
-    def append(self, df, partition_cols: tuple = tuple()):
-        """
-        Executes a simple append operation on a table using the provided dataframe.
-        + Optionally provide a tuple of columns for partitioning the table.
-        """
-
-        return self.create(df, partition_cols)
-
-    def create(self, df, partition_cols: tuple = tuple()):
-        """
-        Executes a simple append operation on a table using the provided dataframe.
-        + Optionally provide a tuple of columns for partitioning the table.
-        """
-        result = (df.write
-                  .format(self.db.table_format())
-                  .partitionBy(partition_cols)
-                  .mode("append")
-                  .saveAsTable(self.db_table_name()))
-        self.merge_table_properties()
-        return result
-
-    @monad.monadic_try(error_cls=error.RepoWriteError)
-    def try_upsert(self, df, partition_puning_col: str = None, partition_cols: tuple = tuple()):
-        """
-        The try_upsert wraps the upsert function with a Try monad.  The result will be an Either.  A successful result
-        usually returns Right(None).
-        """
-        return self.upsert(df, partition_puning_col, partition_cols)
-
-    def upsert(self, df, partition_pruning_col: str = None, partition_cols: tuple = tuple()):
-        """
-        Upsert performs either a create or a delta merge.  The create is called when the table doesnt exist.  Otherwise
-        a delta merge is performed.
-        + partition_pruning_col.  When partitioning, the merge will use one of the partition columns to execute a merge using
-          partition pruning.
-        + Optionally provide partition columns as a tuple of column names.
-
-        Note that the merge requires that the repository implement a identity_merge_condition function when using merge
-        operations.  This is a merge condition that identifies the upsert identity.
-        """
-        if not self.table_exists():
-            return self.create(df, partition_cols)
-
-        (self.delta_table().alias(self.table_name)
-         .merge(
-            df.alias('updates'),
-            self.build_merge_condition(self.table_name, 'updates', partition_pruning_col)
-        )
-         .whenNotMatchedInsertAll()
-         .whenMatchedUpdateAll()
-         .execute())
-
-    #
-    # Streaming Functions
-    #
     def await_termination(self, other_stream_query=None):
         target_stream = other_stream_query if other_stream_query else self.stream_query
         if not target_stream:
@@ -318,12 +386,12 @@ class HiveRepo:
     # Utility Functions
     #
     def build_merge_condition(self, name_of_baseline, update_name, partition_pruning_col):
-        if not hasattr(self, 'identity_merge_condition'):
+        if not self.merge_condition():
             raise repo_messages.error_identity_merge_condition_not_implemented()
 
         pruning_cond = self.build_puning_condition(name_of_baseline, update_name, partition_pruning_col)
 
-        identity_cond = self.identity_merge_condition(name_of_baseline, update_name)
+        identity_cond = self.merge_condition()(name_of_baseline, update_name)
 
         return f"{pruning_cond}{identity_cond}"
 
@@ -357,12 +425,12 @@ class HiveRepo:
     # Table Property Functions
     #
     def merge_table_properties(self):
-        if not hasattr(self, 'table_properties'):
+        if not self.asserted_table_properties():
             return self
         set_on_table = set(self.urn_table_properties())
 
-        self.add_to_table_properties(set(self.__class__.table_properties) - set_on_table)
-        self.remove_from_table_properties(set_on_table - set(self.__class__.table_properties))
+        self.add_to_table_properties(set(self.asserted_table_properties()) - set_on_table)
+        self.remove_from_table_properties(set_on_table - set(self.asserted_table_properties()))
         return self
 
     def urn_table_properties(self) -> List[TableProperty]:
@@ -397,10 +465,9 @@ class DeltaStreamUpserter:
     class TemporaryStreamDumpRepo(HiveRepo):
         table_name = "__temp__"
 
-    def __init__(self, repo: HiveRepo, partition_pruning_col: str = None, partition_cols: tuple = tuple()):
+    def __init__(self, repo: HiveRepo, partition_pruning_col: str = None):
         self.repo = repo
         self.partition_pruning_col = partition_pruning_col
-        self.partition_cols = partition_cols
         self.temp_id = str(uuid.uuid4()).replace("-", "")
         self.temp_repo = self.TemporaryStreamDumpRepo(db=self.repo.db,
                                                       reader=DeltaFileReader,
@@ -408,23 +475,22 @@ class DeltaStreamUpserter:
 
     def execute(self, stream) -> monad.EitherMonad[DataFrame]:
         result = (monad.Right(stream)
-                  >> self.drop_temp_table
+                  >> self.drop_table_by_name
                   >> self.stream_to_temp
                   >> self.await_termination
                   >> self.read_temp
                   >> self.upsert
-                  >> self.drop_temp_table)
+                  >> self.drop_table_by_name)
         return result
 
-    def drop_temp_table(self, stream) -> monad.EitherMonad:
+    def drop_table_by_name(self, stream) -> monad.EitherMonad:
         self.repo.drop_table_by_name(self.temp_table_name(self.temp_repo.table_name))
         return monad.Right(stream)
 
     def stream_to_temp(self, stream) -> monad.EitherMonad:
         self.stream_query = self.temp_repo.write_stream_to_table(stream,
                                                                  self.temp_table_name(self.temp_repo.table_name),
-                                                                 self.repo.__class__.default_stream_trigger_condition,
-                                                                 self.partition_cols)
+                                                                 self.repo.__class__.default_stream_trigger_condition)
         return monad.Right(self.stream_query)
 
     def await_termination(self, stream_query) -> monad.EitherMonad[pyspark.sql.streaming.StreamingQuery]:
@@ -435,7 +501,7 @@ class DeltaStreamUpserter:
         return monad.Right(self.temp_repo.read(self.temp_table_name(self.temp_repo.table_name)))
 
     def upsert(self, df: DataFrame) -> monad.EitherMonad[DataFrame]:
-        self.repo.upsert(df, self.partition_pruning_col, self.partition_cols)
+        self.repo.upsert(df)
         return monad.Right(df)
 
     def temp_table_name(self, temp_prefix):
