@@ -1,6 +1,7 @@
 from typing import Protocol
 from jobsworthy import spark_job
 from jobsworthy.util import error, logger
+from . import repo_messages
 
 
 class DbNamingConventionProtocol(Protocol):
@@ -107,14 +108,15 @@ class DbNamingConventionCallerDefined(DbNamingConventionProtocol):
 
         <dbconfig.db_file_system_path_root>/<database_name>.db
         """
-        if not self.db_file_system_path_root():
-            raise error.RepoConfigError(
-                message="Jobsworth: db path not configured.  Set db_file_system_path_root on the HIVE config")
+        if self.db_file_system_path_root() is None:
+            raise repo_messages.db_path_not_configured()
 
-        if self.db_file_system_path_root()[-1] == "/":
+        if len(self.db_file_system_path_root()) > 1 and self.db_file_system_path_root()[-1] == "/":
             logger.info(msg="Jobsworth: WARNING. db_file_system_path_root should not end with a '/'")
             return f"{self.db_file_system_path_root()[:-1]}/{self.database_name()}.db"
 
+        if not self.db_file_system_path_root():
+            return f"{self.database_name()}.db"
         return f"{self.db_file_system_path_root()}/{self.database_name()}.db"
 
     def db_table_path(self, table_name: str) -> str:
@@ -125,7 +127,6 @@ class DbNamingConventionCallerDefined(DbNamingConventionProtocol):
         """
         return f"{self.db_path()}/{table_name}"
 
-
     def delta_table_location(self, table_name):
         """
         The delta table location (when using the spark context to load a delta table).  Takes into account  delta tables
@@ -158,6 +159,81 @@ class DbNamingConventionCallerDefined(DbNamingConventionProtocol):
             return self.config.db.db_path_override_for_checkpoint
         return self.db_file_system_path_root()
 
+    def checkpoint_location(self, table_name):
+        """
+        The checkpoint location is defined using the delta_table_location (see above) with the constant "_checkpoint"
+        appended.
+        """
+        return f"{self.delta_table_location(table_name)}/{self.checkpoint_folder}"
+
+
+class DbNamingConventionDomainBased(DbNamingConventionProtocol):
+    """
+    DB and Table naming convention based on the names of the domain and data product.  Uses the following properties
+    from the config:
+
+        cfg = (spark_job.JobConfig(data_product_name=my_data_product_name,
+                                   domain_name=my_domain_name
+              .configure_hive_db(db_name="my_db"))
+
+    DB paths ion the cluster are absolute paths (i.e. prepended with a "/".  In test they must be relative paths.
+    This is driven by the setting of job_config().running_in_test().  Therefore, when using this strategy this must
+    be set for testing.
+    """
+
+    checkpoint_folder = "_checkpoint"
+
+    def __init__(self, job_config: spark_job.JobConfig):
+        self.config = job_config
+
+    def database_name(self):
+        return self.config.db.db_name
+
+    def domain_name(self):
+        return self.config.domain_name
+
+    def data_product_name(self):
+        return self.config.data_product_name
+
+    def path_prepend_absolute(self):
+        return "" if self.config.is_running_in_test else "/"
+
+    def test_delta_path_prepend(self):
+        return "spark-warehouse/" if self.config.is_running_in_test else ""
+
+
+    def db_table_name(self, table_name):
+        return f"{self.database_name()}.{table_name}"
+
+    def db_path(self):
+        """
+        Used to define the location of a Hive table.  The DB Path is formatted as follows:
+
+        /domains/<job_config.domain_name>/data_products/<job_config.data_product_name>/<database_name>.db
+        """
+        if not self.domain_name() or not self.data_product_name():
+            raise repo_messages.domain_data_product_not_configured()
+
+        return f"{self.path_prepend_absolute()}domains/{self.domain_name()}/data_products/{self.data_product_name()}/{self.database_name()}.db"
+
+    def db_table_path(self, table_name: str) -> str:
+        """
+        The db_path with the table name appended.  Used when defining the location for a table in the Hive store.
+        :param table_name:
+        :return:
+        """
+        return f"{self.db_path()}/{table_name}"
+
+    def delta_table_location(self, table_name):
+        """
+        The delta table location (when using the spark context to load a delta table).  Takes into account  delta tables
+        with a checkpoint location (when using Delta stream to write data).  On the cluster there is no need to override
+        the checkpoint location (it is db_file_system_path_root from the config).  However, when running the tests the
+        location needs to be overridden by appending "spark-warehouse" to the relative name of the
+        db_file_system_path_root.
+
+        """
+        return f"{self.test_delta_path_prepend()}{self.db_path()}/{table_name}"
 
     def checkpoint_location(self, table_name):
         """
@@ -167,6 +243,7 @@ class DbNamingConventionCallerDefined(DbNamingConventionProtocol):
         return f"{self.delta_table_location(table_name)}/{self.checkpoint_folder}"
 
 
+
 class Db:
 
     def __init__(self, session,
